# Generated by Django 2.0.13 on 2020-05-31 13:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0005_auto_20200516_1726"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="view",
            options={
                "permissions": [
                    ("make_view_public", "Can make a saved view publicly available")
                ]
            },
        ),
        migrations.AlterField(
            model_name="view", name="query", field=models.TextField(blank=True),
        ),
    ]
