# Generated by Django 2.0.13 on 2020-06-14 10:37

from django.db import migrations, models

import data_browser.models


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0006_auto_20200531_1450"),
    ]

    operations = [
        migrations.AddField(
            model_name="view",
            name="public_slug",
            field=models.CharField(default=data_browser.models.get_id, max_length=12),
        ),
    ]
